import base64
import hashlib
import os
import re
from typing import Optional
from logging import getLogger

logger = getLogger(__name__)

def get_alphanumeric_limited_hash(input_str, max_size):
    # Create SHA-256 hash of the input string
    hash_obj = hashlib.sha256(input_str.encode('utf-8'))
    
    # Get the hash digest in base64 format
    hash_base64 = base64.b64encode(hash_obj.digest()).decode('utf-8')
    
    # Remove non-alphanumeric characters and convert to lowercase
    alphanumeric = re.sub(r'[^a-zA-Z0-9]', '', hash_base64).lower()
    
    # Skip the first character to match the Node.js crypto output
    alphanumeric = alphanumeric[1:]
    
    # Limit to max_size characters
    return alphanumeric[:max_size] if len(alphanumeric) > max_size else alphanumeric


def get_global_unique_hash(workspace: str, type: str, name: str) -> str:
    """
    Generate a unique hash for a combination of workspace, type, and name.

    Args:
        workspace: The workspace identifier
        type: The type identifier
        name: The name identifier

    Returns:
        A unique alphanumeric hash string of maximum length 48
    """
    global_unique_name = f"{workspace}-{type}-{name}"
    hash = get_alphanumeric_limited_hash(global_unique_name, 48)
    return hash

class Agent:
    def __init__(self, agent_name: str, workspace: str, run_internal_protocol: str, run_internal_hostname: str):
        self.agent_name = agent_name
        self.workspace = workspace
        self.run_internal_protocol = run_internal_protocol
        self.run_internal_hostname = run_internal_hostname

    @property
    def internal_url(self) -> str:
        """
        Generate the internal URL for the agent using a unique hash.

        Returns:
            The internal URL as a string
        """
        hash_value = get_global_unique_hash(
            self.workspace,
            "agent",
            self.agent_name
        )
        return f"{self.run_internal_protocol}://{hash_value}.{self.run_internal_hostname}"

    @property
    def forced_url(self) -> Optional[str]:
        """
        Check for a forced URL in environment variables.

        Returns:
            The forced URL if found in environment variables, None otherwise
        """
        env_var = self.agent_name.replace("-", "_").upper()
        env_key = f"BL_AGENT_{env_var}_URL"
        return os.environ.get(env_key)



import asyncio

from langgraph.checkpoint.memory import MemorySaver
from langgraph.graph import END, MessagesState, StateGraph
from langgraph.prebuilt import ToolNode
from langgraph.types import Command, interrupt
from typing_extensions import Literal

from blaxel.models import bl_model
from blaxel.tools import bl_tools


class State(MessagesState):
    """Simple state."""

async def create_agent_graph(tools):
    builder = StateGraph(State)

    def human_review_node(state) -> Command[Literal["call_llm", "run_tool"]]:
        last_message = state["messages"][-1]
        tool_call = last_message.tool_calls[-1]

        # this is the value we'll be providing via Command(resume=<human_review>)
        human_review = interrupt(
            {
                "question": "Is this correct?",
                # Surface tool calls for review
                "tool_call": tool_call,
            }
        )

        review_action = human_review["action"]
        review_data = human_review.get("data")

        # if approved, call the tool
        if review_action == "continue":
            return Command(goto="run_tool")

        # update the AI message AND call tools
        elif review_action == "update":
            updated_message = {
                "role": "ai",
                "content": last_message.content,
                "tool_calls": [
                    {
                        "id": tool_call["id"],
                        "name": tool_call["name"],
                        # This the update provided by the human
                        "args": review_data,
                    }
                ],
                # This is important - this needs to be the same as the message you replacing!
                # Otherwise, it will show up as a separate message
                "id": last_message.id,
            }
            return Command(goto="run_tool", update={"messages": [updated_message]})

        # provide feedback to LLM
        elif review_action == "feedback":
            # NOTE: we're adding feedback message as a ToolMessage
            # to preserve the correct order in the message history
            # (AI messages with tool calls need to be followed by tool call messages)
            tool_message = {
                "role": "tool",
                # This is our natural language feedback
                "content": review_data,
                "name": tool_call["name"],
                "tool_call_id": tool_call["id"],
            }
            return Command(goto="call_llm", update={"messages": [tool_message]})

    def route_after_llm(state) -> Literal["__end__", "human_review_node"]:
        if len(state["messages"][-1].tool_calls) == 0:
            return END
        else:
            return "human_review_node"


    async def call_llm(state):
        model = await bl_model("gpt-4o-mini", temperature=0).to_langchain()
        model = model.bind_tools(tools)
        return {"messages": [model.invoke(state["messages"])]}


    builder.add_node("call_llm", call_llm)
    builder.add_node("run_tool", ToolNode(tools))
    builder.add_node("human_review_node", human_review_node)

    builder.set_entry_point("call_llm")
    builder.add_conditional_edges("call_llm", route_after_llm)
    builder.add_edge("run_tool", "call_llm")

    # Set up memory
    memory = MemorySaver()

    # Add
    graph = builder.compile(checkpointer=memory)

    # View
    print(graph.get_graph().draw_ascii())
    return graph



async def run_graphs():
    tools = await bl_tools(["blaxel-search"]).to_langchain()
    graph = await create_agent_graph(tools)

    # With no tool call
    initial_input = {"messages": [{"role": "user", "content": "hi!"}]}
    thread = {"configurable": {"thread_id": "1"}}
    async for event in graph.astream(initial_input, thread, stream_mode="updates"):
        print(event)
        print("\n")


    # With a tool call, should make interupt

    # Run the graph until the first interruption
    initial_input = {"messages": [{"role": "user", "content": "what's the weather in sf?"}]}
    thread = {"configurable": {"thread_id": "2"}}
    async for event in graph.astream(initial_input, thread, stream_mode="updates"):
        print(event)
        print("\n")


    # Pending Executions!
    print("Pending Executions!")
    print(graph.get_state(thread).next)

    # Resume the pending interupt
    async for event in graph.astream(
        # provide value
        Command(resume={"action": "continue"}),
        thread,
        stream_mode="updates",
    ):
        print(event)
        print("\n")

asyncio.run(run_graphs())


from typing import Dict, List

from pydantic import BaseModel


class InstrumentationMapping(BaseModel):
    module_path: str
    class_name: str
    required_packages: List[str]
    ignore_if_packages: List[str]

MAPPINGS: Dict[str, InstrumentationMapping] = {
    "anthropic": InstrumentationMapping(
        module_path="opentelemetry.instrumentation.anthropic",
        class_name="AnthropicInstrumentor",
        required_packages=["anthropic"],
        ignore_if_packages=[]
    ),
    "cohere": InstrumentationMapping(
        module_path="opentelemetry.instrumentation.cohere",
        class_name="CohereInstrumentor",
        required_packages=["cohere"],
        ignore_if_packages=[]
    ),
    "langchain": InstrumentationMapping(
        module_path="opentelemetry.instrumentation.langchain",
        class_name="LangchainInstrumentor",
        required_packages=["langchain"],
        ignore_if_packages=[]
    ),
    "llamaindex": InstrumentationMapping(
        module_path="opentelemetry.instrumentation.llamaindex",
        class_name="LlamaIndexInstrumentor",
        required_packages=["llama_index"],
        ignore_if_packages=[]
    ),
    "crewai": InstrumentationMapping(
        module_path="opentelemetry.instrumentation.crewai",
        class_name="CrewAIInstrumentor",
        required_packages=["crewai"],
        ignore_if_packages=[]
    ),
    "openai": InstrumentationMapping(
        module_path="opentelemetry.instrumentation.openai",
        class_name="OpenAIInstrumentor",
        required_packages=["openai"],
        ignore_if_packages=[]
    ),
}



import os
from typing import Dict

from ..authentication import BlaxelAuth, auth
from .logger import init_logger


class Settings:
    auth: BlaxelAuth

    def __init__(self):
        init_logger(self.log_level)
        self.auth = auth(self.env, self.base_url)
        self._headers = None

    @property
    def env(self) -> str:
        """Get the environment."""
        return os.environ.get("BL_ENV", "prod")

    @property
    def log_level(self) -> str:
        """Get the log level."""
        return os.environ.get("LOG_LEVEL", "INFO").upper()

    @property
    def base_url(self) -> str:
        """Get the base URL for the API."""
        if self.env == "prod":
            return "https://api.blaxel.ai/v0"
        return "https://api.blaxel.dev/v0"

    @property
    def run_url(self) -> str:
        """Get the run URL."""
        if self.env == "prod":
            return "https://run.blaxel.ai"
        return "https://run.blaxel.dev"


    @property
    def headers(self) -> Dict[str, str]:
        """Get the headers for API requests."""
        return self.auth.get_headers()


    @property
    def name(self) -> str:
        """Get the name."""
        return os.environ.get("BL_NAME", "")

    @property
    def type(self) -> str:
        """Get the type."""
        return os.environ.get("BL_TYPE", "agent")

    @property
    def workspace(self) -> str:
        """Get the workspace."""
        return self.auth.workspace_name

    @property
    def run_internal_hostname(self) -> str:
        """Get the run internal hostname."""
        return os.environ.get("BL_RUN_INTERNAL_HOSTNAME", "")

    @property
    def bl_cloud(self) -> bool:
        """Is running on bl cloud."""
        return os.environ.get("BL_CLOUD", "") == "true"

    @property
    def run_internal_protocol(self) -> str:
        """Get the run internal protocol."""
        return os.environ.get("BL_RUN_INTERNAL_PROTOCOL", "https")

    @property
    def enable_opentelemetry(self) -> bool:
        """Get the enable opentelemetry."""
        return os.getenv("BL_ENABLE_OPENTELEMETRY", "false").lower() == "true"

settings = Settings()



from logging import getLogger

from google.adk.models.lite_llm import LiteLlm

from ..common.settings import settings

logger = getLogger(__name__)

async def get_google_adk_model(url: str, type: str, model: str, **kwargs):
    if type == 'mistral':
        return LiteLlm(
            model=f"mistral/{model}",
            api_key=settings.auth.token,
            api_base=f"{url}/v1",
            **kwargs
        )
    elif type == 'cohere':
        return LiteLlm(
            model=f"cohere/{model}",
            api_base=f"{url}/v2/chat",
            extra_headers=settings.auth.get_headers(),
            **kwargs
        )
    elif type == 'xai':
        return LiteLlm(
            model=f"xai/{model}",
            api_key=settings.auth.token,
            api_base=f"{url}/v1",
            **kwargs
        )
    elif type == 'deepseek':
        return LiteLlm(
            model=f"deepseek/{model}",
            api_key=settings.auth.token,
            api_base=f"{url}/v1",
            **kwargs
        )
    elif type == 'anthropic':
        return LiteLlm(
            model=f"anthropic/{model}",
            api_base=url,
            extra_headers=settings.auth.get_headers(),
            **kwargs
        )
    elif type == 'gemini':
        return LiteLlm(
            model=f"gemini/{model}",
            api_base=f"{url}/v1beta/models/{model}",
            extra_headers=settings.auth.get_headers(),
            **kwargs
        )
    elif type == "cerebras":
        return LiteLlm(
            model=f"cerebras/{model}",
            api_key=settings.auth.token,
            api_base=f"{url}/v1",
            **kwargs
        )
    else:
        if type != "openai":
            logger.warning(f"Model {model} is not supported by Google ADK, defaulting to OpenAI")
        return LiteLlm(
            model=f"openai/{model}",
            api_key=settings.auth.token,
            api_base=f"{url}/v1",
            **kwargs
        )


from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="RequestTotalResponseData")


@_attrs_define
class RequestTotalResponseData:
    """Request total response data

    Attributes:
        request_total (Union[Unset, float]): Request total
        status_code (Union[Unset, str]): Status code
        workload_id (Union[Unset, str]): Workload ID
        workload_type (Union[Unset, str]): Workload type
        workspace (Union[Unset, str]): Workspace
    """

    request_total: Union[Unset, float] = UNSET
    status_code: Union[Unset, str] = UNSET
    workload_id: Union[Unset, str] = UNSET
    workload_type: Union[Unset, str] = UNSET
    workspace: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        request_total = self.request_total

        status_code = self.status_code

        workload_id = self.workload_id

        workload_type = self.workload_type

        workspace = self.workspace

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if request_total is not UNSET:
            field_dict["requestTotal"] = request_total
        if status_code is not UNSET:
            field_dict["statusCode"] = status_code
        if workload_id is not UNSET:
            field_dict["workloadId"] = workload_id
        if workload_type is not UNSET:
            field_dict["workloadType"] = workload_type
        if workspace is not UNSET:
            field_dict["workspace"] = workspace

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        request_total = d.pop("requestTotal", UNSET)

        status_code = d.pop("statusCode", UNSET)

        workload_id = d.pop("workloadId", UNSET)

        workload_type = d.pop("workloadType", UNSET)

        workspace = d.pop("workspace", UNSET)

        request_total_response_data = cls(
            request_total=request_total,
            status_code=status_code,
            workload_id=workload_id,
            workload_type=workload_type,
            workspace=workspace,
        )

        request_total_response_data.additional_properties = d
        return request_total_response_data

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties




import httpx
from httpx import AsyncClient
from pydantic_ai.providers import Provider


class GoogleGLAProvider(Provider[AsyncClient]):
    """Provider for Google Generative Language AI API."""

    @property
    def name(self):
        return 'google-gla'

    @property
    def base_url(self) -> str:
        return 'https://generativelanguage.googleapis.com/v1beta/models/'

    @property
    def client(self) -> httpx.AsyncClient:
        return self._client

    def __init__(self, api_key, http_client: httpx.AsyncClient | None = None) -> None:
        """Create a new Google GLA provider.

        Args:
            api_key: The API key to use for authentication, if not provided, the `GEMINI_API_KEY` environment variable
                will be used if available.
            http_client: An existing `httpx.AsyncClient` to use for making HTTP requests.
        """
        self._client = http_client
        # https://cloud.google.com/docs/authentication/api-keys-use#using-with-rest
        self._client.headers['X-Goog-Api-Key'] = api_key



import asyncio
import json
import os
from contextlib import AsyncExitStack
from logging import getLogger
from typing import Any, cast
import traceback

from mcp import ClientSession
from mcp.types import CallToolResult
from mcp.types import Tool as MCPTool

from ..common.env import env
from ..common.internal import get_global_unique_hash
from ..common.settings import settings
from ..instrumentation.span import SpanManager
from ..mcp.client import websocket_client
from .types import Tool

logger = getLogger(__name__)

DEFAULT_TIMEOUT = 1
if os.getenv("BL_SERVER_PORT"):
    DEFAULT_TIMEOUT = 5

class PersistentWebSocket:
    def __init__(self, url: str, name: str, timeout: int = DEFAULT_TIMEOUT, timeout_enabled: bool = True):
        self.url = url
        self.name = name
        self.timeout = timeout
        self.session_exit_stack = AsyncExitStack()
        self.client_exit_stack = AsyncExitStack()
        self.session: ClientSession = None
        self.timer_task = None
        self.tools_cache = []
        if settings.bl_cloud:
            self.timeout_enabled = False
        else:
            self.timeout_enabled = timeout_enabled

    def with_metas(self, metas: dict[str, Any]):
        self.metas = metas
        return self

    async def call_tool(self, tool_name: str, arguments: dict[str, Any]) -> CallToolResult:
        await self._initialize()
        if self.timeout_enabled:
            self._remove_timer()
        logger.debug(f"Calling tool {tool_name} with arguments {arguments}")
        arguments.update(self.metas)
        call_tool_result = await self.session.call_tool(tool_name, arguments)
        logger.debug(f"Tool {tool_name} returned {call_tool_result}")
        if self.timeout_enabled:
            self._reset_timer()
        else:
            await self._close()
        return call_tool_result

    async def list_tools(self):
        logger.debug(f"Listing tools for {self.name}")
        span_attributes = {
            "tool.server": self.url,
            "tool.server_name": self.name,
            "span.type": "tool.list",
        }
        with SpanManager("blaxel-tracer").create_active_span(self.name, span_attributes) as span:
            await self._initialize()
            logger.debug(f"Initialized websocket for {self.name}")
            if self.timeout_enabled:
                self._remove_timer()
            logger.debug("Listing tools")
            list_tools_result = await self.session.list_tools()
            self.tools_cache = list_tools_result.tools
            logger.debug(f"Tools listed: {list_tools_result}")
            if self.timeout_enabled:
                self._reset_timer()
            else:
                await self._close()
            span.set_attribute("tool.list.result", list_tools_result.model_dump_json())
            return list_tools_result

    def get_tools(self):
        return self.tools_cache

    async def _initialize(self):
        if not self.session:
            logger.debug(f"Initializing websocket client for {self.url}")
            read, write = await self.client_exit_stack.enter_async_context(websocket_client(self.url, settings.headers))
            self.session = cast(ClientSession, await self.session_exit_stack.enter_async_context(ClientSession(read, write)))
            await self.session.initialize()

    def _reset_timer(self):
        self._remove_timer()
        self.timer_task = asyncio.create_task(self._close_after_timeout())

    def _remove_timer(self):
        if self.timer_task:
            self.timer_task.cancel()

    async def _close_after_timeout(self):
        await asyncio.sleep(self.timeout)
        await self._close()
        self.session = None


    async def _close(self):
        logger.debug(f"Closing websocket client {self.url}")
        if self.session:
            self.session = None
            try:
                await self.session_exit_stack.aclose()
            except Exception as e:
                logger.debug(f"Error closing session exit stack: {e}")
            try:
                await self.client_exit_stack.aclose()
            except Exception as e:
                logger.debug(f"Error closing client exit stack: {e}")
            logger.debug("WebSocket connection closed due to inactivity.")

def convert_mcp_tool_to_blaxel_tool(
    websocket_client: PersistentWebSocket,
    name: str,
    url: str,
    tool: MCPTool,
) -> Tool:
    """Convert an MCP tool to a blaxel tool.

    NOTE: this tool can be executed only in a context of an active MCP client session.

    Args:
        session: MCP client session
        tool: MCP tool to convert

    Returns:
        a LangChain tool
    """

    async def initialize_and_call_tool(
        *args: Any,
        **arguments: dict[str, Any],
    ) -> CallToolResult:
        span_attributes = {
            "tool.name": tool.name,
            "tool.args": json.dumps(arguments),
            "tool.server": url,
            "tool.server_name": name,
            "span.type": "tool.call",
        }
        with SpanManager("blaxel-tracer").create_active_span("blaxel-tool-call", span_attributes):
            logger.debug(f"Calling tool {tool.name} with arguments {arguments}")
            call_tool_result = await websocket_client.call_tool(tool.name, arguments)
            logger.debug(f"Tool {tool.name} returned {call_tool_result}")
            return call_tool_result

    async def call_tool(
        *args: Any,
        **arguments: dict[str, Any],
    ) -> CallToolResult:
        return await initialize_and_call_tool(*args, **arguments)

    def sync_call_tool(*args: Any, **arguments: dict[str, Any]) -> CallToolResult:
        try:
            loop = asyncio.get_running_loop()
            return loop.run_until_complete(initialize_and_call_tool(*args, **arguments))
        except RuntimeError:
            return asyncio.run(initialize_and_call_tool(*args, **arguments))

    return Tool(
        name=tool.name,
        description=tool.description or "",
        input_schema=tool.inputSchema,
        coroutine=call_tool,
        sync_coroutine=sync_call_tool,
        response_format="content_and_artifact",
    )

toolPersistances: dict[str, PersistentWebSocket] = {}

class BlTools:
    def __init__(self, functions: list[str], metas: dict[str, Any] = {}, timeout: int = DEFAULT_TIMEOUT, timeout_enabled: bool = True):
        self.functions = functions
        self.metas = metas
        self.timeout = timeout
        self.timeout_enabled = timeout_enabled

    def _internal_url(self, name: str):
        """Get the internal URL for the agent using a hash of workspace and agent name."""
        hash = get_global_unique_hash(settings.workspace, "function", name)
        return f"{settings.run_internal_protocol}://bl-{settings.env}-{hash}.{settings.run_internal_hostname}"

    def _forced_url(self, name: str):
        """Get the forced URL from environment variables if set."""
        env_var = name.replace("-", "_").upper()
        if env[f"BL_FUNCTION_{env_var}_URL"]:
            return env[f"BL_FUNCTION_{env_var}_URL"]
        return None

    def _external_url(self, name: str):
        return f"{settings.run_url}/{settings.workspace}/functions/{name}"

    def _fallback_url(self, name: str):
        if self._external_url(name) != self._url(name):
            return self._external_url(name)
        return None

    def _url(self, name: str):
        logger.debug(f"Getting URL for {name}")
        if self._forced_url(name):
            logger.debug(f"Forced URL found for {name}: {self._forced_url(name)}")
            return self._forced_url(name)
        if settings.run_internal_hostname:
            logger.debug(f"Internal hostname found for {name}: {self._internal_url(name)}")
            return self._internal_url(name)
        logger.debug(f"No URL found for {name}, using external URL")
        return self._external_url(name)

    def get_tools(self) -> list[Tool]:
        """Get a list of all tools from all connected servers."""
        all_tools: list[Tool] = []
        for name in self.functions:
            toolPersistances.get(name).with_metas(self.metas)
            websocket = toolPersistances.get(name)
            tools = websocket.get_tools()
            converted_tools = [convert_mcp_tool_to_blaxel_tool(websocket, name, self._url(name), tool) for tool in tools]
            all_tools.extend(converted_tools)
        return all_tools

    async def to_langchain(self):
        from .langchain import get_langchain_tools

        await self.initialize()
        return get_langchain_tools(self.get_tools())

    async def to_llamaindex(self):
        from .llamaindex import get_llamaindex_tools

        await self.initialize()
        return get_llamaindex_tools(self.get_tools())

    async def to_crewai(self):
        from .crewai import get_crewai_tools

        await self.initialize()
        return get_crewai_tools(self.get_tools())

    async def to_openai(self):
        from .openai import get_openai_tools

        await self.initialize()
        return get_openai_tools(self.get_tools())

    async def to_pydantic(self):
        from .pydantic import get_pydantic_tools

        await self.initialize()
        return get_pydantic_tools(self.get_tools())

    async def to_google_adk(self):
        from .googleadk import get_google_adk_tools

        await self.initialize()
        return get_google_adk_tools(self.get_tools())

    async def to_livekit(self):
        from .livekit import get_livekit_tools

        await self.initialize()
        return get_livekit_tools(self.get_tools())

    async def connect(self, name: str):
        # Create and store the connection
        try:
            url = self._url(name)
            await self.connect_with_url(name, url)
        except Exception as e:
            if not self._fallback_url(name):
                raise e
            logger.warning(f"Error connecting to {name}: {e}\n{traceback.format_exc()}")
            url = self._fallback_url(name)
            try:
                await self.connect_with_url(name, url)
            except Exception as e:
                logger.error(f"Error connecting to {name} with fallback URL: {e}\n{traceback.format_exc()}")
                raise e

    async def connect_with_url(
        self, name: str, url: str
    ) -> None:
        """Initialize a session and load tools from it.

        Args:
            name: Name to identify this server connection
            url: The URL to connect to
        """
        logger.debug(f"Initializing session and loading tools from {url}")

        if not toolPersistances.get(name):
            logger.debug(f"Creating new persistent websocket for {name}")
            toolPersistances[name] = PersistentWebSocket(url, name, timeout=self.timeout, timeout_enabled=self.timeout_enabled)
            await toolPersistances[name].list_tools()
        logger.debug(f"Loaded {len(toolPersistances[name].get_tools())} tools from {url}")
        return toolPersistances[name].with_metas(self.metas)

    async def initialize(self) -> "BlTools":
        for i in range(0, len(self.functions), 10):
            batch = self.functions[i:i+10]
            await asyncio.gather(*(self.connect(name) for name in batch))
        return self


def bl_tools(functions: list[str], metas: dict[str, Any] = {}, timeout: int = DEFAULT_TIMEOUT, timeout_enabled: bool = True) -> BlTools:
    return BlTools(functions, metas=metas, timeout=timeout, timeout_enabled=timeout_enabled)



import json
from logging import getLogger
from typing import Any, Awaitable

from ..cache import find_from_cache
from ..client import client
from ..client.api.agents import get_agent
from ..client.models import Agent
from ..common.env import env
from ..common.internal import get_global_unique_hash
from ..common.settings import settings
from ..instrumentation.span import SpanManager

logger = getLogger(__name__)

class BlAgent:
    def __init__(self, name: str):
        self.name = name

    @property
    def internal_url(self):
        """Get the internal URL for the agent using a hash of workspace and agent name."""
        hash = get_global_unique_hash(settings.workspace, "agent", self.name)
        return f"{settings.run_internal_protocol}://bl-{settings.env}-{hash}.{settings.run_internal_hostname}"

    @property
    def forced_url(self):
        """Get the forced URL from environment variables if set."""
        env_var = self.name.replace("-", "_").upper()
        if env[f"BL_AGENT_{env_var}_URL"]:
            return env[f"BL_AGENT_{env_var}_URL"]
        return None

    @property
    def external_url(self):
        return f"{settings.run_url}/{settings.workspace}/agents/{self.name}"

    @property
    def fallback_url(self):
        if self.external_url != self.url:
            return self.external_url
        return None

    @property
    def url(self):
        if self.forced_url:
            return self.forced_url
        if settings.run_internal_hostname:
            return self.internal_url
        return self.external_url

    def call(self, url, input_data, headers: dict = {}, params: dict = {}):
        body = input_data
        if not isinstance(body, str):
            body = json.dumps(body)

        return client.get_httpx_client().post(
            url,
            headers={
                'Content-Type': 'application/json',
                **headers
            },
            data=body,
            params=params
        )

    async def acall(self, url, input_data, headers: dict = {}, params: dict = {}):
        logger.debug(f"Agent Calling: {self.name}")
        body = input_data
        if not isinstance(body, str):
            body = json.dumps(body)

        return await client.get_async_httpx_client().post(
            url,
            headers={
                'Content-Type': 'application/json',
                **headers
            },
            data=body,
            params=params
        )

    def run(self, input: Any, headers: dict = {}, params: dict = {}) -> str:
        attributes = {
            "agent.name": self.name,
            "agent.args": json.dumps(input),
            "span.type": "agent.run",
        }
        with SpanManager("blaxel-tracer").create_active_span(self.name, attributes) as span:
            logger.debug(f"Agent Calling: {self.name}")
            response = self.call(self.url, input, headers, params)
            if response.status_code >= 400:
                if not self.fallback_url:
                    span.set_attribute("agent.run.error", response.text)
                    raise Exception(f"Agent {self.name} returned status code {response.status_code} with body {response.text}")
                response = self.call(self.fallback_url, input, headers, params)
                if response.status_code >= 400:
                    span.set_attribute("agent.run.error", response.text)
                    raise Exception(f"Agent {self.name} returned status code {response.status_code} with body {response.text}")
            span.set_attribute("agent.run.result", response.text)
            return response.text

    async def arun(self, input: Any, headers: dict = {}, params: dict = {}) -> Awaitable[str]:
        attributes = {
            "agent.name": self.name,
            "agent.args": json.dumps(input),
            "span.type": "agent.run",
        }
        with SpanManager("blaxel-tracer").create_active_span(self.name, attributes) as span:
            logger.debug(f"Agent Calling: {self.name}")
            response = await self.acall(self.url, input, headers, params)
            if response.status_code >= 400:
                if not self.fallback_url:
                    span.set_attribute("agent.run.error", response.text)
                    raise Exception(f"Agent {self.name} returned status code {response.status_code} with body {response.text}")
                response = await self.acall(self.fallback_url, input, headers, params)
                if response.status_code >= 400:
                    span.set_attribute("agent.run.error", response.text)
                    raise Exception(f"Agent {self.name} returned status code {response.status_code} with body {response.text}")
            span.set_attribute("agent.run.result", response.text)
            return response.text

    def __str__(self):
        return f"Agent {self.name}"

    def __repr__(self):
        return self.__str__()

def bl_agent(name: str):
    return BlAgent(name)

async def get_agent_metadata(name):
    cache_data = await find_from_cache('Agent', name)
    if cache_data:
        return Agent.from_dict(cache_data)
    try:
        return await get_agent.asyncio(client=client, agent_name=name)
    except Exception:
        return None



from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.serverless_config_configuration import ServerlessConfigConfiguration


T = TypeVar("T", bound="ServerlessConfig")


@_attrs_define
class ServerlessConfig:
    """Configuration for a serverless deployment

    Attributes:
        configuration (Union[Unset, ServerlessConfigConfiguration]): The configuration for the deployment
        max_scale (Union[Unset, int]): The minimum number of replicas for the deployment. Can be 0 or 1 (in which case
            the deployment is always running in at least one location).
        min_scale (Union[Unset, int]): The maximum number of replicas for the deployment.
        timeout (Union[Unset, int]): The timeout for the deployment in seconds
    """

    configuration: Union[Unset, "ServerlessConfigConfiguration"] = UNSET
    max_scale: Union[Unset, int] = UNSET
    min_scale: Union[Unset, int] = UNSET
    timeout: Union[Unset, int] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        configuration: Union[Unset, dict[str, Any]] = UNSET
        if (
            self.configuration
            and not isinstance(self.configuration, Unset)
            and not isinstance(self.configuration, dict)
        ):
            configuration = self.configuration.to_dict()
        elif self.configuration and isinstance(self.configuration, dict):
            configuration = self.configuration

        max_scale = self.max_scale

        min_scale = self.min_scale

        timeout = self.timeout

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if configuration is not UNSET:
            field_dict["configuration"] = configuration
        if max_scale is not UNSET:
            field_dict["maxScale"] = max_scale
        if min_scale is not UNSET:
            field_dict["minScale"] = min_scale
        if timeout is not UNSET:
            field_dict["timeout"] = timeout

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.serverless_config_configuration import ServerlessConfigConfiguration

        if not src_dict:
            return None
        d = src_dict.copy()
        _configuration = d.pop("configuration", UNSET)
        configuration: Union[Unset, ServerlessConfigConfiguration]
        if isinstance(_configuration, Unset):
            configuration = UNSET
        else:
            configuration = ServerlessConfigConfiguration.from_dict(_configuration)

        max_scale = d.pop("maxScale", UNSET)

        min_scale = d.pop("minScale", UNSET)

        timeout = d.pop("timeout", UNSET)

        serverless_config = cls(
            configuration=configuration,
            max_scale=max_scale,
            min_scale=min_scale,
            timeout=timeout,
        )

        serverless_config.additional_properties = d
        return serverless_config

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties
